/* 
 *  Vnc2swf - writeswf.c (generates a SWF file with Ming)
 *  $Id: writeswf.c,v 1.5 2004/05/19 16:02:01 yusuke Exp $
 *
 *  Copyright (C) 2002 by Yusuke Shinyama (yusuke at cs . nyu . edu)
 *  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
*/


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>

#include <X11/Xaw/Toggle.h>

/* ming */
#include <ming.h>
#include <zlib.h>

#include "vncviewer.h"


#define DEBUG_SWF 0
#define DEBUG_CLIP 0
#define DEBUG_DBL 0
#define DEBUG_ZBUFFER 0

#define CHANNELS 4

static char* movie_filename;
static FILE* movie_dumpf;
static int movie_width, movie_height;

static SWFMovie movie;
static FILE* soundf;

static unsigned long tmpbuf_len;
static char* tmpbuf_src;
static char* tmpbuf_compressed;
static long intervalusec;
static struct timeval tv0;
static struct timezone dummytz;

static int img_xmin, img_ymin, img_xmax, img_ymax;

static void reset_pending_image(void);
static void track_image(int x, int y, int w, int h);
static void flush_pending_image(void);


/* ZBuffer: covered pixels = 1 */
typedef unsigned short zpixel;
typedef struct _DisplayItem {
  SWFDisplayItem d;
  zpixel id;
  int x, y, w, h;
  struct _DisplayItem* next;
  struct _DisplayItem* prev;
} DisplayItem;

static zpixel shapeid;
static zpixel* zbuffer;
static DisplayItem items;

static int trigger;


/* initializeZBuffer:
 */
static Bool initializeZBuffer(void)
{
    zpixel* p;
    unsigned long buflen;
    buflen = (unsigned long)movie_width*(unsigned long)movie_height*sizeof(zpixel);
    p = zbuffer = (zpixel*)malloc(buflen);
    if (!zbuffer) {
	perror("initializeZBuffer: Out of memory");
	return False;
    }
    items.d = NULL;
    items.next = &items;
    items.prev = &items;
    return True;
}

static void clearZBuffer(void)
{
    DisplayItem* di = items.next;

    memset(zbuffer, 0, (unsigned long)movie_width*(unsigned long)movie_height*sizeof(zpixel));
    shapeid = 1;
    while(di->d) {
	DisplayItem* removed = di;
	di->next->prev = di->prev;
	di->prev->next = di->next;
	di = di->next;
	free((char*)removed);	
    }
}


/* paintZBuffer:
 *   Fill the given rect with the current shape id.
 */
static Bool paintZBuffer(SWFDisplayItem d, int x, int y, int w, int h)
{
    zpixel* p = zbuffer + y * movie_width + x;
    DisplayItem* di;
    
#if DEBUG_ZBUFFER
    fprintf(stderr,"paintZBuffer: %u: (%d,%d)-(%d,%d)\n",
	    shapeid, x, y, x+w, y+h);
#endif
    di = (DisplayItem*)malloc(sizeof(DisplayItem));
    if (!di) {
	perror("fillZBuffer: Out of memory");
	return False;
    }
    di->d = d;
    di->id = shapeid;
    di->x = x;
    di->y = y;
    di->w = w;
    di->h = h;
    /* WARNING: h is broken after this! */
    while(h--) {
	zpixel* p1 = p;
	int w1 = w;
	while(w1--) {
	    *(p1++) = shapeid;
	}
	p += movie_width;
    }
    
    /* add to the list */
    di->next = items.next;
    di->prev = &items;
    items.next->prev = di;
    items.next = di;
    shapeid++;
    return True;
}


/* checkZBufferCovered:
 *   Check if every pixel in the given rect is covered thereafter.
 */
static Bool checkZBufferCovered(zpixel id, int x, int y, int w, int h)
{
    zpixel* p = zbuffer + y * movie_width + x;
    /* WARNING: h is broken after this! */
    while(h--) {
	zpixel* p1 = p;
	int w1 = w;
	while(w1--) {
	    if (*(p1++) <= id) return False; /* this pixel is not yet covered! */
	}
	p += movie_width; 
    }
    return True;
}


/* checkRemoveItems:
 *   Remove covered items from the movie.
 */
static void checkRemoveItems(void)
{
    DisplayItem* di = items.next;
    
    while(di->d) {
	if (checkZBufferCovered(di->id, di->x, di->y, di->w, di->h)) {
	    DisplayItem* removed = di;
#if DEBUG_ZBUFFER
	    fprintf(stderr,"checkRemoveItems: covered: %u: (%d,%d)-(%d,%d)\n",
		    di->id, di->x, di->y, di->x+di->w, di->y+di->h);
#endif
	    SWFMovie_remove(movie, di->d);
	    
	    /* remove from the list */
	    di->next->prev = di->prev;
	    di->prev->next = di->next;
	    di = di->next;
	    free((char*)removed);
	} else {
	    di = di->next;
	}
    }
}


/* ConvBufToRGB:
 *   Convert the area (x,y),(w,h) in a framebuffer (ZPixmap) to a RGB array.
 */
static void ConvBufToRGB(unsigned char* buf, int skip, int w, int h)
{
    int i, j;
    char* p = tmpbuf_src;
    
    for(i = 0; i < h; i++) {
	unsigned char* linebuf = buf;
	for(j = 0; j < w; j++) {
	    unsigned long c = 0, c1, c2, c3;
	    switch(myFormat.bitsPerPixel) {
	    case 8:
		if (appData.useBGR233) {
		    c = BGR233ToPixel[*(linebuf++)];
		} else {
		    c = *(linebuf++);
		}
		break;
	    case 16:
		c1 = *(linebuf++);
		c = (c1<<8) | *(linebuf++);
		c = Swap16IfLE(c);
		break;
	    case 32:
		c1 = *(linebuf++);
		c2 = *(linebuf++);
		c3 = *(linebuf++);
		c = (c1<<24) | (c2<<16) | (c3<<8) | *(linebuf++);
		c = Swap32IfLE(c);
		break;
	    }
	    /* RGB */
	    *(p++) = 0;
	    *(p++) = redRv[(c>>myFormat.redShift) & myFormat.redMax]; /* R */
	    *(p++) = greenRv[(c>>myFormat.greenShift) & myFormat.greenMax]; /* G */
	    *(p++) = blueRv[(c>>myFormat.blueShift) & myFormat.blueMax]; /* B */
	}
	buf += skip;
    }
}


/* ImageToDBL:
 *   Make a DBL image from raw framebuffers.
 *
 * DBL format:
 *   3bytes: 'DBl' (header)
 *   1byte : format (1:noalpha, 2:hasalpha)
 *   4bytes: size of the data contained below (n+6 or n+5)
 *   1byte : palette? (5:nopalette, 3:haspalette)
 *   2bytes: width
 *   2bytes: height
 *   [1byte: num_palette-1] (if has a palette)
 *   nbytes: compressed RGB or RGBA
 */
/* write_ulong */
static char* write_ulong(char* p, unsigned long x)
{
    *(p++) = (x >> 24) & 0xff;
    *(p++) = (x >> 16) & 0xff;
    *(p++) = (x >> 8) & 0xff;
    *(p++) = (x >> 0) & 0xff;
    return p;
}
/* write_ushort */
static char* write_ushort(char* p, unsigned short x)
{
    *(p++) = (x >> 0) & 0xff;
    *(p++) = (x >> 8) & 0xff;
    return p;
}
static char* ImageToDBL(unsigned long* len, int x, int y, int w, int h)
{
    int scrWidthInBytes;
    char* buf;
    char* dblbuf;
    char* p;
    unsigned long dbllen, dbldatalen, dbltotallen;
    unsigned long imglen = (unsigned long)w * (unsigned long)h * CHANNELS;
    
    scrWidthInBytes = si.framebufferWidth * myFormat.bitsPerPixel / 8;
    buf = image->data + y * scrWidthInBytes + x * myFormat.bitsPerPixel / 8;
    ConvBufToRGB(buf, scrWidthInBytes, w, h);
    
    dbllen = tmpbuf_len;
    compress(tmpbuf_compressed, &dbllen, tmpbuf_src, imglen);
    dbldatalen = dbllen+1+2+2;
    dbltotallen = dbldatalen+3+1+4;
    dblbuf = malloc(dbltotallen);
    if (!dblbuf) return(NULL);
    appData.compressed_image_bytes += dbllen;
    appData.expanded_image_bytes += imglen;
    
    p = dblbuf;
    /* a DBL header */
    *(p++) = 'D';
    *(p++) = 'B';
    *(p++) = 'l';
    /* no alpha */
    *(p++) = 1;
    /* the dummy data size */
    p = write_ulong(p, dbldatalen);
    /* no pallete */
    *(p++) = 5;			
    /* the image size */
    p = write_ushort(p, w);
    p = write_ushort(p, h);
    /* the compressed data */
    memcpy(p, tmpbuf_compressed, dbllen);
    
    *len = dbltotallen;
    
#if DEBUG_DBL
    {
	/* dump each DBL file into files 
	 * CAUTION: files can be too big and many! 
	 */
	char fname[256];
	FILE* fp; 
	sprintf(fname, "f%05u.dbl", shapeid);
	fp = fopen(fname, "wb");
	fwrite(dblbuf, 1, dbltotallen, fp);
	fclose(fp);
    }
#endif  
    
    return(dblbuf);
}


/*  Lower level swf operations.
 *
 */

/* WriteRawImage:
 *   Put an image obtained from a buffer (x0,y0) into (x1,y1) in the movie.
 */
static Bool WriteRawImage(int x0, int y0, int w, int h, int x1, int y1)
{
    SWFShape shape;
    SWFFill fill;
    SWFDBLBitmap dbl;
    SWFDisplayItem d;
    SWFInput input;
    char* dblbuf;
    unsigned long dbllen;
    
#if DEBUG_SWF
    fprintf(stderr,"WriteRawImage: %dx%d (%d, %d) to (%d, %d)\n", w, h, x0, y0, x1, y1);
#endif /* DEBUG_SWF */
    
    dblbuf = ImageToDBL(&dbllen, x0, y0, w, h);
    if (!dblbuf) return(False);
    input = newSWFInput_allocedBuffer(dblbuf, dbllen);
    dbl = newSWFDBLBitmap_fromInput(input);
    shape = newSWFShape();
    fill = SWFShape_addBitmapFill(shape, (SWFBitmap)dbl, SWFFILL_TILED_BITMAP);
    SWFShape_setRightFill(shape, fill);
    if (appData.macromediaHack) {
	SWFShape_movePenTo(shape, 1, 1);
    } else {
	SWFShape_movePenTo(shape, 0, 0);
    }
    SWFShape_drawLine(shape, w, 0);
    SWFShape_drawLine(shape, 0, h);
    SWFShape_drawLine(shape,-w, 0);
    SWFShape_drawLine(shape, 0,-h);
    SWFShape_end(shape);
    d = SWFMovie_add(movie, (SWFBlock)shape);
    SWFDisplayItem_moveTo(d, x1, y1);
    
    paintZBuffer(d, x1, y1, w, h);
    checkRemoveItems();
    return True;
}


/* WriteRawRectangle:
 *   Put a rectangle.
 */
static Bool WriteRawRectangle(Pixel c, int x, int y, int w, int h)
{
    SWFShape shape;
    SWFFill fill;
    SWFDisplayItem d;
    int red, green, blue;
    
#if DEBUG_SWF
    fprintf(stderr,"WriteRawRectangle: %dx%d to (%d, %d)\n", w, h, x, y);
#endif /* DEBUG_SWF */
    
    shape = newSWFShape();
    red = redRv[(c>>myFormat.redShift) & myFormat.redMax];
    green = greenRv[(c>>myFormat.greenShift) & myFormat.greenMax];
    blue = blueRv[(c>>myFormat.blueShift) & myFormat.blueMax];
    fill = SWFShape_addSolidFill(shape, red, green, blue, 255);
    SWFShape_setRightFill(shape, fill);
    if (appData.macromediaHack) {
	SWFShape_movePenTo(shape, 1, 1);
    } else {
	SWFShape_movePenTo(shape, 0, 0);
    }
    SWFShape_drawLine(shape, w, 0);
    SWFShape_drawLine(shape, 0, h);
    SWFShape_drawLine(shape, -w, 0);
    SWFShape_drawLine(shape, 0, -h);
    d = SWFMovie_add(movie, (SWFBlock)shape);
    SWFDisplayItem_moveTo(d, x, y);
    
    paintZBuffer(d, x, y, w, h);
    checkRemoveItems();
    return True;
}


/* checkNextFrame:
 *   Add frames necessary.
 */
static void checkNextFrame(void)
{
    struct timeval tv;
    
    gettimeofday(&tv, &dummytz);
    if (tv0.tv_sec < 0) {		/* initialize */
	tv0 = tv;
	tv0.tv_usec += intervalusec;
	if (1000000L <= tv0.tv_usec) {
	    tv0.tv_usec -= 1000000L;
	    tv0.tv_sec++;
	}
    }
    while(tv0.tv_sec < tv.tv_sec || 
	  (tv0.tv_sec == tv.tv_sec && tv0.tv_usec < tv.tv_usec)) {
	if (appData.recordingMethod == RECORD_BUFFERED) {
	    flush_pending_image();
	}
	SWFMovie_nextFrame(movie);
	appData.nframes++;
#if DEBUG_SWF
	fprintf(stderr, "checkNextFrame: frame: %d\n", appData.nframes);
#endif /* DEBUG_SWF */
	tv0.tv_usec += intervalusec;
	if (1000000L <= tv0.tv_usec) {
	    tv0.tv_usec -= 1000000L;
	    tv0.tv_sec++;
	}    
    }
}

/* WriteInitMovie:
 *   Initialize Ming and prepare a movie file.
 */
static void disp_status(void)
{
    fprintf(stderr, "=== ");
    if (appData.recording) {
	ChangeToplevelTitle("[Recording] ");
	fprintf(stderr, "Start Recording");
	switch(appData.recordingMethod) {
	case RECORD_IMMEDIATELY:
	    fprintf(stderr, " (RECORD_IMMEDIATELY)");
	    break;
	case RECORD_BUFFERED:
	    fprintf(stderr, " (RECORD_BUFFERED)");
	    break;
	}
    } else {
	ChangeToplevelTitle("[Stopped] ");
	fprintf(stderr, "Stop Recording");
    }
    fprintf(stderr, "\n");
}

void
SwitchRecording(void)
{
    if (appData.recording) {
	gettimeofday(&tv0, &dummytz);
	flush_pending_image();
    }
    disp_status();
    trigger = False;
}

Bool init_recording(void)
{
    if (soundf != NULL) {
	fclose(soundf);
    }
    if (movie != NULL) {
	destroySWFMovie(movie);
    }
    
    if (appData.soundfileString) {
	soundf = fopen(appData.soundfileString, "rb");
	if (!soundf) {
	    perror("WriteInitMovie: Cannot open sound file");
	    return False;
	}
    }

    movie = newSWFMovie();
    SWFMovie_setRate(movie, appData.frameRate);
    if (soundf) {
	fprintf(stderr,"=== WriteInitMovie: Sound file: \"%s\"\n", appData.soundfileString);
#if MING03
	/* ming 0.3beta1 */
	SWFMovie_setSoundStream(movie, (SWFSoundStream)newSWFSoundStream(soundf));
#else
	/* ming 0.2a */
	SWFMovie_setSoundStream(movie, newSWFSound(soundf));
#endif
    }
    SWFMovie_setDimension(movie, movie_width, movie_height);
    intervalusec = 1000000L / appData.frameRate;
    tv0.tv_sec = tv0.tv_usec = -1;
    appData.nframes = 0;
    trigger = False;
    
    appData.recording = appData.startRecording || (!appData.showDesktop);
    appData.compressed_image_bytes = 0;
    appData.expanded_image_bytes = 0;
    
    clearZBuffer();
    track_image(0, 0, movie_width, movie_height);

    disp_status();
    return True;
}

Bool WriteInitMovie(int w, int h)
{
    Ming_init();

    movie_filename = appData.dumpfileString;
    movie_width = w;
    movie_height = h;

    {
	/* warn if the file is going to be overwritten */
	struct stat st;
	if (stat(movie_filename, &st) != -1) {
	    fprintf(stderr, "=== WriteInitMovie: Warning: file %s already exists, overwriting.\n", 
		    appData.dumpfileString);
	}
    }
    fprintf(stderr,"=== WriteInitMovie: Pid=%d, Opening file: \"%s\" for a movie size (%d, %d), frame rate %d...\n", 
	    getpid(), movie_filename, movie_width, movie_height, appData.frameRate);
    
    movie_dumpf = fopen(movie_filename, "wb");
    if (!movie_dumpf) {
	perror("WriteInitMovie: Cannot open file");
	return False;
    }
    
    tmpbuf_len = (unsigned long)floor((unsigned long)w*(unsigned long)h*CHANNELS*1.01+12);
    tmpbuf_src = malloc(tmpbuf_len);
    tmpbuf_compressed = malloc(tmpbuf_len);
    if (!tmpbuf_src || !tmpbuf_compressed) {
	perror("WriteInitMovie: Out of memory");
	return False;
    }
    initializeZBuffer();
    
    movie = NULL;
    soundf = NULL;
    return init_recording();
}


/* WriteFinishMovie:
 *   Write the movie file.
 */
Bool WriteFinishMovie(void)
{
    long len;
    
    /* flush */
    if (appData.recording) {
	flush_pending_image();
	checkNextFrame();
    }
    
    fprintf(stderr, "=== WriteFinishMovie: %d frames (%.1f sec), expanded_image: %lu\n",
	    appData.nframes, (double)appData.nframes/(double)appData.frameRate, appData.expanded_image_bytes);

#if MING02
    /* ming 0.3beta1 */
    len = SWFMovie_output(movie, fileOutputMethod, movie_dumpf, 0);
#else
    /* ming 0.2a */
    len = SWFMovie_output(movie, fileOutputMethod, movie_dumpf);
#endif
    destroySWFMovie(movie);
    
    if (soundf != NULL) {
	fclose(soundf);
    }
    
    if (0 < len) {
	fprintf(stderr, "=== WriteFinishMovie: %ld bytes dumped into \"%s\".\n", len, movie_filename);
    } else {
	perror("=== WriteFinishMovie: cannot write the file!");
	return False;
    }
    return True;
}


/* return true if the whole rectangle is clipped. */
int cliprect(int* x0, int* y0, int* w, int* h, int* x1, int* y1)
{
    if (*x0 < appData.cgx0) {
	int dx = appData.cgx0 - *x0;
	*w -= dx;
	*x0 += dx;
	*x1 += dx;
    } else if (appData.cgx1 <= *x0) {
	return 1;
    }

    if (*y0 < appData.cgy0) {
	int dy = appData.cgy0 - *y0;
	*h -= dy;
	*y0 += dy;
	*y1 += dy;
    } else if (appData.cgy1 <= *y0) {
	return 1;
    }

    if (appData.cgx1 < *x1+*w) {
	*w = appData.cgx1 - *x1;
    }
    if (appData.cgy1 < *y1+*h) {
	*h = appData.cgy1 - *y1;
    }
    *x1 -= appData.cgx0;
    *y1 -= appData.cgy0;
#if DEBUG_CLIP
    fprintf(stderr, "%dx%d (%d,%d) -> (%d,%d)\n", *w, *h, *x0, *y0, *x1, *y1);
#endif
    return (*w <= 0) || (*h <= 0);
}

/* WriteImage:
 */
Bool WriteImage(int x0, int y0, int w, int h)
{
    int x1 = x0, y1 = y0;
#if DEBUG_SWF
    fprintf(stderr,"WriteImage: %dx%d from (%d, %d)\n", w, h, x0, y0);
#endif /* DEBUG_SWF */

    /* check if a signal is sent. */
    if (trigger) {
	SwitchRecording();
    }
    if (!appData.recording) {
	track_image(x0, y0, w, h);
	return True;
    }
    
    checkNextFrame();
    if (appData.recordingMethod == RECORD_BUFFERED) {
	track_image(x0, y0, w, h);
	return True;
    }
    if (cliprect(&x0, &y0, &w, &h, &x1, &y1)) return True;

    return WriteRawImage(x0, y0, w, h, x1, y1);
}

/* WriteRectangle:
 */
Bool WriteRectangle(Pixel c, int x0, int y0, int w, int h)
{
    int x1 = x0, y1 = y0;
#if DEBUG_SWF
    fprintf(stderr,"WriteRectangle: %dx%d from (%d, %d)\n", w, h, x0, y0);
#endif /* DEBUG_SWF */

    /* check if a signal is sent. */
    if (trigger) {
	SwitchRecording();
    }
    if (!appData.recording) {
	track_image(x0, y0, w, h);
	return True;
    }
    
    checkNextFrame();
    if (appData.recordingMethod == RECORD_BUFFERED) {
	track_image(x0, y0, w, h);
	return True;
    }
    if (cliprect(&x0, &y0, &w, &h, &x1, &y1)) return True;

    return WriteRawRectangle(c, x1, y1, w, h);
}

/* WriteCopyImage:
 *   Copy the image.
 */
Bool WriteCopyImage(Display* dpy, Drawable d, 
		    int x0, int y0, int w, int h, int x1, int y1)
{
#if DEBUG_SWF
    fprintf(stderr,"WriteCopyImage: %dx%d from (%d, %d) to (%d, %d)\n", w, h, x0, y0, x1, y1);
#endif /* DEBUG_SWF */
    /* check if a signal is sent. */
    if (trigger) {
	SwitchRecording();
    }
    if (!appData.recording) {
	track_image(x1, y1, w, h);
	return True;
    }
  
    checkNextFrame();
    if (appData.recordingMethod == RECORD_BUFFERED) {
	track_image(x1, y1, w, h);
	return True;
    }
    
    if (cliprect(&x0, &y0, &w, &h, &x1, &y1)) return True;
    
    /* destroy the content of the variable "image" */
    XGetSubImage(dpy, d, x0, y0, w, h, 0xffffffff, ZPixmap, image, x0, y0);
    
    return WriteRawImage(x0, y0, w, h, x1, y1);
}


/*
 * track pixels which have not been recorded. 
 */

#define INF 9999
static void reset_pending_image(void) {
    img_xmin = img_ymin = INF;
    img_xmax = img_ymax =-INF;
}

static void track_image(int x, int y, int w, int h) {
    if (x < img_xmin) img_xmin = x;
    if (y < img_ymin) img_ymin = y;
    if (img_xmax < x+w) img_xmax = x+w;
    if (img_ymax < y+h) img_ymax = y+h;
#if DEBUG_SWF
    fprintf(stderr,"track_image: now (%d, %d)-(%d, %d)\n", img_xmin, img_ymin, img_xmax, img_ymax);
#endif /* DEBUG_SWF */
}

static void flush_pending_image(void) {
    if (img_xmin <= img_xmax && img_ymin <= img_ymax) {
	int x0 = img_xmin, y0 = img_ymin;
	int x1 = x0, y1 = y0;
	int w = img_xmax-img_xmin, h = img_ymax-img_ymin;
	if (cliprect(&x0, &y0, &w, &h, &x1, &y1)) return;
	WriteRawImage(x0, y0, w, h, x1, y1);
#if DEBUG_SWF
	fprintf(stderr,"flush_pending_image: (%d, %d)-(%d, %d)\n", img_xmin, img_ymin, img_xmax, img_ymax);
#endif /* DEBUG_SWF */
	reset_pending_image();
    }
}

void
SetRecordingState(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    XtVaSetValues(w, XtNstate, appData.recording, NULL);
}

void
ToggleRecording(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    appData.recording = !appData.recording;
    SwitchRecording();
    SetRecordingState(w, event, 0, 0);
}

void
SwitchMethod(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    appData.recordingMethod = (appData.recordingMethod+1) % 2;
    fprintf(stderr, "=== SwitchMethod: ");
    switch(appData.recordingMethod) {
    case RECORD_IMMEDIATELY:
	flush_pending_image();
	fprintf(stderr, "RECORD_IMMEDIATELY\n");
	break;
    case RECORD_BUFFERED:
	fprintf(stderr, "RECORD_BUFFERED\n");
	break;
    default:
	break;
    }
}

void
ResetRecording(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    fprintf(stderr, "=== Reset Recording.\n");
    init_recording();    
}


/* signal handler
 */
void TurnRecordingOn(int sig)
{
    fprintf(stderr,"TurnRecordingOn called\n");
    appData.recording = True;
    trigger = True;
}
void TurnRecordingOff(int sig)
{
    fprintf(stderr,"TurnRecordingOff called\n");
    appData.recording = False;
    trigger = True;
}
